import asyncio
import dataclasses
import logging
from typing import *

import numpy as np

from hippolyzer.lib.base.datatypes import UUID, Vector3, Vector2
from hippolyzer.lib.base.message.message import Message, Block
from hippolyzer.lib.base.templates import ParcelGridFlags, ParcelFlags
from hippolyzer.lib.client.state import BaseClientRegion


LOG = logging.getLogger(__name__)


@dataclasses.dataclass
class Parcel:
    local_id: int
    name: str
    flags: ParcelFlags
    group_id: UUID
    # TODO: More properties


class ParcelManager:
    # We expect to receive this number of ParcelOverlay messages
    NUM_CHUNKS = 4
    # No, we don't support varregion or whatever.
    REGION_SIZE = 256
    # Basically, the minimum parcel size is 4 on either axis so each "point" in the
    # ParcelOverlay represents an area this size
    GRID_STEP = 4
    GRIDS_PER_EDGE = REGION_SIZE // GRID_STEP

    def __init__(self, region: BaseClientRegion):
        # dimensions are south to north, west to east
        self.overlay = np.zeros((self.GRIDS_PER_EDGE, self.GRIDS_PER_EDGE), dtype=np.uint8)
        # 1-indexed parcel list index
        self.parcel_indices = np.zeros((self.GRIDS_PER_EDGE, self.GRIDS_PER_EDGE), dtype=np.uint16)
        self.parcels: List[Optional[Parcel]] = []
        self.overlay_chunks: List[Optional[bytes]] = [None] * self.NUM_CHUNKS
        self.overlay_complete = asyncio.Event()
        self.parcels_downloaded = asyncio.Event()
        self._parcels_dirty: bool = True
        self._region = region
        self._next_seq = 1
        self._region.message_handler.subscribe("ParcelOverlay", self._handle_parcel_overlay)

    def _handle_parcel_overlay(self, message: Message):
        self.add_overlay_chunk(message["ParcelData"]["Data"], message["ParcelData"]["SequenceID"])

    def add_overlay_chunk(self, chunk: bytes, chunk_num: int) -> bool:
        self.overlay_chunks[chunk_num] = chunk
        # Still have some pending chunks, don't try to parse this yet
        if not all(self.overlay_chunks):
            return False

        new_overlay_data = b"".join(self.overlay_chunks)
        self.overlay_chunks = [None] * self.NUM_CHUNKS
        self._parcels_dirty = False
        if new_overlay_data != self.overlay.data[:]:
            # If the raw data doesn't match, then we have to parse again
            new_data = np.frombuffer(new_overlay_data, dtype=np.uint8).reshape(self.overlay.shape)
            np.copyto(self.overlay, new_data)
            self._parse_overlay()
            # We could optimize this by just marking specific squares dirty
            # if the parcel indices have changed between parses, but I don't care
            # to do that.
            self._parcels_dirty = True
            self.parcels_downloaded.clear()
        if not self.overlay_complete.is_set():
            self.overlay_complete.set()
        return True

    @classmethod
    def _pos_to_grid_coords(cls, pos: Vector3) -> Tuple[int, int]:
        return round(pos.Y // cls.GRID_STEP), round(pos.X // cls.GRID_STEP)

    def _parse_overlay(self):
        # Zero out all parcel indices
        self.parcel_indices[:, :] = 0
        next_parcel_idx = 1
        for y in range(0, self.GRIDS_PER_EDGE):
            for x in range(0, self.GRIDS_PER_EDGE):
                # We already have a parcel index for this grid, continue
                if self.parcel_indices[y, x]:
                    continue

                # Fill all adjacent grids with this parcel index
                self._flood_fill_parcel_index(y, x, next_parcel_idx)
                # SL doesn't allow disjoint grids to be part of the same parcel, so
                # whatever grid we find next without a parcel index must be a new parcel
                next_parcel_idx += 1

        # Should have found at least one parcel
        assert next_parcel_idx >= 2

        # Have a different number of parcels now, we can't use the existing parcel objects
        # because it's unlikely that just parcel boundaries have changed.
        if len(self.parcels) != next_parcel_idx - 1:
            # We don't know about any of these parcels yet, fill with none
            self.parcels = [None] * (next_parcel_idx - 1)

    def _flood_fill_parcel_index(self, start_y, start_x, parcel_idx):
        """Flood fill all neighboring grids with the parcel index, being mindful of parcel boundaries"""
        # We know the start grid is assigned to this parcel index
        self.parcel_indices[start_y, start_x] = parcel_idx
        # Queue of grids to test the neighbors of, start with the start grid.
        neighbor_test_queue: List[Tuple[int, int]] = [(start_y, start_x)]

        while neighbor_test_queue:
            to_test = neighbor_test_queue.pop(0)
            test_grid = self.overlay[to_test]

            for direction in ((-1, 0), (1, 0), (0, -1), (0, 1)):
                new_pos = to_test[0] + direction[0], to_test[1] + direction[1]

                if any(x < 0 or x >= self.GRIDS_PER_EDGE for x in new_pos):
                    # Outside bounds
                    continue
                if self.parcel_indices[new_pos]:
                    # Already set, skip
                    continue

                if direction[0] == -1 and test_grid & ParcelGridFlags.SOUTH_LINE:
                    # Test grid is already on a south line, can't go south.
                    continue
                if direction[1] == -1 and test_grid & ParcelGridFlags.WEST_LINE:
                    # Test grid is already on a west line, can't go west.
                    continue

                grid = self.overlay[new_pos]

                if direction[0] == 1 and grid & ParcelGridFlags.SOUTH_LINE:
                    # Hit a south line going north, this is outside the current parcel
                    continue
                if direction[1] == 1 and grid & ParcelGridFlags.WEST_LINE:
                    # Hit a west line going east, this is outside the current parcel
                    continue
                # This grid is within the current parcel, set the parcel index
                self.parcel_indices[new_pos] = parcel_idx
                # Append the grid to the neighbour testing queue
                neighbor_test_queue.append(new_pos)

    async def request_dirty_parcels(self) -> Tuple[Parcel, ...]:
        if self._parcels_dirty:
            return await self.request_all_parcels()
        return tuple(self.parcels)

    async def request_all_parcels(self) -> Tuple[Parcel, ...]:
        await self.overlay_complete.wait()
        # Because of how we build up the parcel index map, it's safe for us to
        # do this instead of keeping track of seen IDs in a set or similar
        last_seen_parcel_index = 0
        futs = []
        for y in range(0, self.GRIDS_PER_EDGE):
            for x in range(0, self.GRIDS_PER_EDGE):
                parcel_index = self.parcel_indices[y, x]
                assert parcel_index != 0
                if parcel_index <= last_seen_parcel_index:
                    continue
                assert parcel_index == last_seen_parcel_index + 1
                last_seen_parcel_index = parcel_index
                # Request a position within the parcel
                futs.append(self.request_parcel_properties(
                    Vector2(x * self.GRID_STEP + 1.0, y * self.GRID_STEP + 1.0)
                ))

        # Wait for all parcel properties to come in
        await asyncio.gather(*futs)
        self.parcels_downloaded.set()
        self._parcels_dirty = False
        return tuple(self.parcels)

    async def request_parcel_properties(self, pos: Vector2) -> Parcel:
        await self.overlay_complete.wait()
        seq_id = self._next_seq
        # Register a wait on a ParcelProperties matching this seq
        parcel_props_fut = self._region.message_handler.wait_for(
            ("ParcelProperties",),
            predicate=lambda msg: msg["ParcelData"]["SequenceID"] == seq_id,
            timeout=10.0,
        )
        # We don't care about when we receive an ack, we only care about when we receive the parcel props
        _ = self._region.circuit.send_reliable(Message(
            "ParcelPropertiesRequest",
            Block("AgentData", AgentID=self._region.session().agent_id, SessionID=self._region.session().id),
            Block(
                "ParcelData",
                SequenceID=seq_id,
                West=pos.X,
                East=pos.X,
                North=pos.Y,
                South=pos.Y,
                # What does this even mean?
                SnapSelection=0,
            ),
        ))
        self._next_seq += 1

        return self._process_parcel_properties(await parcel_props_fut, pos)

    def _process_parcel_properties(self, parcel_props: Message, pos: Optional[Vector2] = None) -> Parcel:
        data_block = parcel_props["ParcelData"][0]
        grid_coord = None
        # Parcel indices are one-indexed, convert to zero-indexed.
        if pos is not None:
            # We have a pos, figure out where in the grid we should look for the parcel index
            grid_coord = self._pos_to_grid_coords(pos)
        else:
            # Need to look at the parcel bitmap to figure out a valid grid coord.
            # This is a boolean array where each bit says whether the parcel occupies that grid.
            parcel_bitmap = data_block.deserialize_var("Bitmap")

            for y in range(self.GRIDS_PER_EDGE):
                for x in range(self.GRIDS_PER_EDGE):
                    if parcel_bitmap[y, x]:
                        # This is the first grid the parcel occupies per the bitmap
                        grid_coord = y, x
                        break
                if grid_coord:
                    break

        parcel = Parcel(
            local_id=data_block["LocalID"],
            name=data_block["Name"],
            flags=ParcelFlags(data_block["ParcelFlags"]),
            group_id=data_block["GroupID"],
            # Parcel UUID isn't in this response :/
        )

        # I guess the bitmap _could_ be empty, but probably not.
        if grid_coord is not None:
            parcel_idx = self.parcel_indices[grid_coord] - 1
            if len(self.parcels) > parcel_idx >= 0:
                # Okay, parcels list is sane, place the parcel in there.
                self.parcels[parcel_idx] = parcel
            else:
                LOG.warning(f"Received ParcelProperties with incomplete overlay for {grid_coord!r}")

        return parcel

    async def get_parcel_at(self, pos: Vector2, request_if_missing: bool = True) -> Optional[Parcel]:
        grid_coord = self._pos_to_grid_coords(pos)
        parcel = None
        if parcel_idx := self.parcel_indices[grid_coord]:
            parcel = self.parcels[parcel_idx - 1]
        if request_if_missing and parcel is None:
            return await self.request_parcel_properties(pos)
        return parcel



# TODO: some fancy parser that parses everything into
#  dicts or objects using schemas.
from __future__ import annotations

import asyncio
import weakref
from typing import Any, Optional, Coroutine, NamedTuple

import defusedxml.lxml
import lxml.etree


class VivoxMessage(NamedTuple):
    type: str
    name: str
    request_id: Optional[str]
    data: dict


def xml_to_dict(element):
    return element.tag, dict(map(xml_to_dict, element)) or element.text


def buildxml(r, d, list_elem_name='i'):
    if isinstance(d, dict):
        for k, v in d.items():
            s = lxml.etree.SubElement(r, k)
            buildxml(s, v, list_elem_name)
    elif isinstance(d, (list, tuple, set)):
        for v in d:
            if isinstance(v, lxml.etree._Element):  # noqa
                s = r
            else:
                s = lxml.etree.SubElement(r, list_elem_name)
            buildxml(s, v, list_elem_name)
    elif isinstance(d, str):
        r.text = d
    elif isinstance(d, lxml.etree._Element):  # noqa
        r.append(d)
    elif d is None:
        r.text = ""
    else:
        r.text = str(d)
    return r


_VIVOX_NS = b' xmlns="http://www.vivox.com"'  # noqa


def _remove_vivox_ns(data):
    return data.replace(_VIVOX_NS, b"").strip()


def _clean_message(msg_action: str, parsed, dict_msg: dict):
    # TODO: refactor this into some XML -> dict schema, some XML is ambiguous
    if msg_action == "Aux.GetCaptureDevices.1":
        devices = []
        for device in parsed.find('Results/CaptureDevices'):
            devices.append(xml_to_dict(device)[1])
        dict_msg["Results"]["CaptureDevices"] = devices
    if msg_action == "Account.WebCall.1":
        results = dict_msg["Results"]
        content_type = results.get("ContentType") or ""
        if content_type.startswith("text/xml"):
            xml_content = _remove_vivox_ns(results["Content"].encode("utf8"))
            parsed_content = defusedxml.lxml.fromstring(xml_content)
            body = parsed_content.xpath("//body")[0]
            results["Content"] = body
    if "ReturnCode" in dict_msg:
        dict_msg["ReturnCode"] = int(dict_msg["ReturnCode"])
    return dict_msg


def _build_webcall_params(params: dict) -> list:
    params_list = []
    elem = lxml.etree.Element('base')
    for name, val in params.items():
        params_list.append({"Name": name, "Value": val})
    buildxml(elem, params_list, 'Parameter')
    return list(elem)


class VivoxConnection:
    def __init__(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter, owned=True):
        self._reader: Optional[asyncio.StreamReader] = reader
        self._writer: Optional[asyncio.StreamWriter] = writer
        self._owned = owned

    def close(self):
        if self._owned and self._writer:
            self._writer.close()
        self._writer = None
        self._reader = None

    def __del__(self):
        self.close()

    async def read_messages(self):
        # TODO: handle interrupted read
        while self._reader and not self._reader.at_eof() and not self._writer.is_closing():
            yield await self.read_message()

    async def read_message(self):
        msg = await self._reader.readuntil(b"\n\n\n")
        return self.parse(msg[:-3])

    def parse(self, raw_msg) -> VivoxMessage:
        parsed_msg = defusedxml.lxml.fromstring(raw_msg.decode("utf8"))
        msg_type = parsed_msg.tag
        request_id = parsed_msg.attrib.get("requestId", None)

        # There may be no params, just use an empty dict if that's the case
        dict_msg = xml_to_dict(parsed_msg)[1] or {}

        if msg_type == "Event":
            msg_action = parsed_msg.attrib.get("type")
        elif msg_type == "Response":
            msg_action = parsed_msg.attrib.get("action")
            # This is pretty useless, get rid of it because it gunks up repr()s.
            if 'InputXml' in dict_msg:
                del dict_msg['InputXml']
            dict_msg = _clean_message(msg_action, parsed_msg, dict_msg)
        elif msg_type == "Request":
            msg_action = parsed_msg.attrib.get("action")
        else:
            raise Exception("Unknown Vivox message type %r?" % msg_type)
        return VivoxMessage(msg_type, msg_action, request_id, dict_msg)

    def send_raw(self, buf: bytes) -> Coroutine[Any, Any, None]:
        self._writer.write(buf + b"\n\n\n")
        drain_coro = self._writer.drain()
        # Don't whine if this isn't awaited, we may not always want to flush immediately.
        weakref.finalize(drain_coro, drain_coro.close)
        return drain_coro

    def send_request(self, request_id: str, action: str, data: Any) -> Coroutine[Any, Any, None]:
        if action == "Account.WebCall.1":
            data = dict(data)
            data["Parameters"] = _build_webcall_params(data["Parameters"])
        return self._send_request_response("Request", request_id, action, data)

    def send_response(self, request_id: str, action: str, data: Any) -> Coroutine[Any, Any, None]:
        return self._send_request_response("Response", request_id, action, data)

    def _send_request_response(self, msg_type: str, request_id: str, action: str, data: Any):
        elem = lxml.etree.Element(msg_type)
        elem.attrib["requestId"] = request_id
        elem.attrib["action"] = action
        serialized = lxml.etree.tostring(buildxml(elem, data))
        return self.send_raw(serialized)

    def send_event(self, event_type: str, data: Any) -> Coroutine[Any, Any, None]:
        elem = lxml.etree.Element("Event")
        elem.attrib["type"] = event_type
        serialized = lxml.etree.tostring(buildxml(elem, data))
        return self.send_raw(serialized)



from __future__ import annotations

import asyncio
import base64
import json
import logging
import random
import subprocess
import tempfile
import urllib.parse
import uuid
from typing import Optional, Union, Any, Dict

from hippolyzer.lib.base.datatypes import Vector3
from hippolyzer.lib.base.events import Event
from hippolyzer.lib.base.message.message_handler import MessageHandler
from hippolyzer.lib.base.objects import handle_to_gridxy
from .connection import VivoxConnection, VivoxMessage

LOG = logging.getLogger(__name__)
RESP_LOG = logging.getLogger(__name__ + ".responses")


def launch_slvoice(voice_path, args, env=None):
    return subprocess.Popen([voice_path] + args, env=env)


def uuid_to_vivox(val):
    return (b"x" + base64.b64encode(uuid.UUID(val).bytes, b"-_")).decode("utf8")


def uuid_to_vivox_uri(val):
    return "sip:%s@bhr.vivox.com" % uuid_to_vivox(val)


def vivox_to_uuid(val):
    # Pull the base64-encoded UUID out of the URI
    val = val.split(":")[-1].split("@")[0][1:]
    return str(uuid.UUID(bytes=base64.b64decode(val, b"-_")))


class VoiceClient:
    SERVER_URL = "https://www.bhr.vivox.com/api2/"

    def __init__(self, host: str, port: int):
        self._host = host
        self._port = port

        self.logged_in = asyncio.Event()
        self.ready = asyncio.Event()
        self.session_ready = asyncio.Event()
        self.session_added = Event()
        self.channel_info_updated = Event()
        self.participant_added = Event()
        self.participant_updated = Event()
        self.participant_removed = Event()
        self.capture_devices_received = Event()
        self.render_devices_received = Event()
        self.render_devices = {}
        self.capture_devices = {}

        self._pending_req_futures: dict[str, asyncio.Future] = {}

        self._connector_handle: Optional[str] = None
        self._session_handle: Optional[str] = None
        self._session_group_handle: Optional[str] = None
        self._account_handle: Optional[str] = None
        self._account_uri: Optional[str] = None
        self._username: Optional[str] = None
        self._password: Optional[str] = None
        self._display_name: Optional[str] = None
        self._uri: Optional[str] = None
        self._participants: Dict[str, dict] = {}

        self._mic_muted = False
        self._region_global_x = 0
        self._region_global_y = 0

        self._pos = Vector3(0, 0, 0)

        self.vivox_conn: Optional[VivoxConnection] = None
        self._poll_task = asyncio.create_task(self._poll_messages())
        self.event_handler: MessageHandler[VivoxMessage, str] = MessageHandler(take_by_default=False)

        self.event_handler.subscribe(
            "VoiceServiceConnectionStateChangedEvent",
            self._handle_voice_service_connection_state_changed
        )
        self.event_handler.subscribe("AccountLoginStateChangeEvent", self._handle_account_login_state_change)
        self.event_handler.subscribe("SessionAddedEvent", self._handle_session_added)
        self.event_handler.subscribe("SessionRemovedEvent", self._handle_session_removed)
        self.event_handler.subscribe("ParticipantAddedEvent", self._handle_participant_added)
        self.event_handler.subscribe("ParticipantUpdatedEvent", self._handle_participant_updated)
        self.event_handler.subscribe("ParticipantRemovedEvent", self._handle_participant_removed)

    @property
    def username(self):
        return self._username

    @property
    def password(self):
        return self._password

    @property
    def display_name(self):
        return self._display_name

    @property
    def global_pos(self):
        return self._pos

    @property
    def region_pos(self):
        return self._global_to_region(self.global_pos)

    @property
    def uri(self):
        return self._uri

    @property
    def participants(self):
        # TODO: wrap in something to make immutable
        return self._participants

    def close(self):
        if self.vivox_conn is not None:
            self.vivox_conn.close()
        self._poll_task.cancel()
        self._poll_task = None

    async def aclose(self):
        if self._account_handle:
            await self.logout()
        self.close()

    @classmethod
    async def simple_init(
            cls,
            voice_path: str,
            host: Optional[str] = None,
            port: Optional[int] = None,
            env: Optional[dict] = None
    ):
        """Simple initializer for standing up a client"""
        if not host:
            host = "127.0.0.1"
        if not port:
            port = random.randrange(40000, 60000)

        str_addr = "%s:%s" % (host, port)
        launch_slvoice(voice_path, ["-i", str_addr, "-m", "component"], env=env)
        # HACK: wait for the process to start listening
        await asyncio.sleep(0.2)

        client = cls(host, port)
        await client.create_vivox_connection()
        await client.ready.wait()
        return client

    async def create_vivox_connection(self):
        reader, writer = await asyncio.open_connection(host=self._host, port=self._port)
        self.vivox_conn = VivoxConnection(reader, writer)

    async def create_connector(self):
        # TODO: Move all this extra crap out of here
        devices = (await self.send_message("Aux.GetCaptureDevices.1", {}))["Results"]
        self.capture_devices_received.notify(devices)
        self.capture_devices.clear()
        self.capture_devices.update(devices)

        devices = (await self.send_message("Aux.GetRenderDevices.1", {}))["Results"]
        self.render_devices_received.notify(devices)
        self.render_devices.clear()
        self.render_devices.update(devices)

        await self.set_speakers_muted(False)
        await self.set_speaker_volume(62)
        await self.set_mic_muted(True)
        await self.set_mic_volume(50)

        connector_resp = await self.send_message("Connector.Create.1", {
            "ClientName": "V2 SDK",
            "AccountManagementServer": self.SERVER_URL,
            "Mode": "Normal",
            "MinimumPort": 30000,
            "MaximumPort": 50000,
            "Logging": {
                "Folder": tempfile.gettempdir(),
                "FileNamePrefix": "VivConnector",
                "FileNameSuffix": ".log",
                "LogLevel": 1
            },
            "Application": "",
            "MaxCalls": 12,
        })

        self._connector_handle = connector_resp['Results']['ConnectorHandle']
        self.ready.set()

    async def login(self, username: Union[uuid.UUID, str], password: str):
        # UUID, convert to Vivox format
        if isinstance(username, uuid.UUID) or len(username) == 36:
            username = uuid_to_vivox(username)

        self._username = username
        self._password = password
        if not self._connector_handle:
            raise Exception("Need a connector handle to log in")
        if self._account_handle:
            await self.logout()

        resp = await self.send_message("Account.Login.1", {
            "ConnectorHandle": self._connector_handle,
            "AccountName": username,
            "AccountPassword": password,
            "AudioSessionAnswerMode": "VerifyAnswer",
            "EnableBuddiesAndPresence": "false",
            "BuddyManagementMode": "Application",
            "ParticipantPropertyFrequency": 5,
        })

        if resp["ReturnCode"] != 0:
            raise Exception(resp)

        self._display_name = urllib.parse.unquote(resp["Results"]["DisplayName"])
        self._account_uri = resp["Results"]["Uri"]
        await self.logged_in.wait()

        return resp

    async def logout(self):
        if self._session_handle:
            await self.leave_session()

        if self._account_handle:
            await self.send_message("Account.Logout.1", {
                "AccountHandle": self._account_handle,
            })
            self._account_handle = None
            self._account_uri = None
            self.logged_in.clear()

    async def join_session(self, uri: str, region_handle: Optional[int] = None):
        if self._session_handle:
            await self.leave_session()

        self.set_ref_region(region_handle)

        self._uri = uri

        await self.send_message("Session.Create.1", {
            "AccountHandle": self._account_handle,
            "URI": uri,
            "ConnectAudio": "true",
            "ConnectText": "false",
            "VoiceFontID": 0,
            "Name": ""
        })
        # wait until we're actually added
        await self.session_ready.wait()

    async def leave_session(self):
        await self.send_message("SessionGroup.Terminate.1", {
            "SessionGroupHandle": self._session_group_handle,
        })
        self.session_ready.clear()

        # TODO: refactor into a collection
        for participant in self._participants.values():
            self.participant_removed.notify(participant)
        self._participants.clear()
        self._session_handle = None
        self._session_group_handle = None
        self._region_global_x = 0
        self._region_global_y = 0
        self._uri = None

    def set_3d_pos(self, pos: Vector3, vel: Vector3 = Vector3(0, 0, 0)) -> asyncio.Future:
        """Set global 3D position, in Vivox coordinates"""
        self._pos = pos
        future = self.send_message("Session.Set3DPosition.1", {
            "SessionHandle": self._session_handle,
            "SpeakerPosition": self._build_position_dict(pos),
            "ListenerPosition": self._build_position_dict(pos, vel=vel),
        })
        self._channel_info_updated()
        return future

    def set_region_3d_pos(self, pos: Vector3, vel: Vector3 = Vector3(0, 0, 0)) -> asyncio.Future:
        """Set 3D position, in region-local coordinates"""
        vel = Vector3(vel[0], vel[2], -vel[1])
        return self.set_3d_pos(self._region_to_global(pos), vel=vel)

    def set_speakers_muted(self, val: bool):
        return self.send_message("Connector.MuteLocalSpeaker.1", {
            "Value": json.dumps(val),
            "ConnectorHandle": self._connector_handle
        })

    def set_mic_muted(self, val: bool):
        self._mic_muted = val

        return self.send_message("Connector.MuteLocalMic.1", {
            "Value": json.dumps(val),
            "ConnectorHandle": self._connector_handle
        })

    def set_mic_volume(self, vol: int):
        return self.send_message("Connector.SetLocalMicVolume.1", {
            "Value": vol,
            "ConnectorHandle": self._connector_handle
        })

    def set_speaker_volume(self, vol: int):
        return self.send_message("Connector.SetLocalSpeakerVolume.1", {
            "Value": vol,
            "ConnectorHandle": self._connector_handle
        })

    def set_capture_device(self, device: str):
        return self.send_message("Aux.SetCaptureDevice.1", {
            "CaptureDeviceSpecifier": device,
        })

    def set_participant_volume(self, participant: str, vol: int):
        return self.send_message("Session.SetParticipantVolumeForMe.1", {
            "SessionHandle": self._session_handle,
            "ParticipantURI": participant,
            "Volume": vol,
        })

    async def get_channel_info(self, uri: str) -> dict:
        return await self.send_message("Account.ChannelGetInfo.1", {
            "AccountHandle": self._account_handle,
            "URI": uri
        })

    def send_web_call(self, rel_path: str, params: dict) -> asyncio.Future[dict]:
        """Make a call to a Vivox Web API"""
        return self.send_message("Account.WebCall.1", {
            "AccountHandle": self._account_handle,
            "RelativePath": rel_path,
            "Parameters": params,
        })

    def send_message(self, msg_type: str, data: Any) -> asyncio.Future[dict]:
        request_id = self._make_request_id()
        # This is apparently what the viewer does, not clear if
        # request_id has any semantic significance
        if msg_type == "Session.Create.1":
            request_id = data["URI"]

        RESP_LOG.debug("%s %s %s %r" % ("Request", request_id, msg_type, data))

        asyncio.create_task(self.vivox_conn.send_request(request_id, msg_type, data))
        future = asyncio.Future()
        self._pending_req_futures[request_id] = future
        return future

    def send_raw(self, data: bytes):
        return self.vivox_conn.send_raw(data)

    def set_ref_region(self, region_handle: Optional[int]):
        """Set reference position for region-local coordinates"""
        if region_handle is not None:
            self._region_global_x, self._region_global_y = handle_to_gridxy(region_handle)
        else:
            self._region_global_x, self._region_global_y = (0, 0)
        self._channel_info_updated()

    async def _poll_messages(self):
        while not self.vivox_conn:
            await asyncio.sleep(0.001)

        async for msg in self.vivox_conn.read_messages():
            try:
                RESP_LOG.debug(repr(msg))
                if msg.type == "Event":
                    self.event_handler.handle(msg)
                elif msg.type == "Response":
                    # Might not have this request ID if it was sent directly via the socket
                    if msg.request_id in self._pending_req_futures:
                        self._pending_req_futures[msg.request_id].set_result(msg.data)
                        del self._pending_req_futures[msg.request_id]
            except Exception:
                LOG.exception("Error in response handler?")

    async def _handle_voice_service_connection_state_changed(self, _msg: VivoxMessage):
        await self.create_connector()

    def _handle_account_login_state_change(self, msg: VivoxMessage):
        if msg.data.get('StatusString') == "OK" and msg.data['State'] == '1':
            self._account_handle = msg.data['AccountHandle']
            self.logged_in.set()
        else:
            self.logged_in.clear()
            self._account_uri = None
            self._account_handle = None

    def _handle_session_added(self, msg: VivoxMessage):
        self._session_handle = msg.data["SessionHandle"]
        self._session_group_handle = msg.data["SessionGroupHandle"]
        self.session_added.notify(self._session_handle)
        # We still have to wait for ourselves to be added as a participant, wait on
        # that to set the session_ready event.

    def _handle_session_removed(self, _msg: VivoxMessage):
        self._session_handle = None
        # We often don't get all the `ParticipantRemoved`s before the session dies,
        # clear out the participant list.
        for participant in tuple(self._participants.keys()):
            self._remove_participant(participant)
        self.session_ready.clear()

    def _handle_participant_added(self, msg: VivoxMessage):
        self._participants[msg.data["ParticipantUri"]] = msg.data
        self.participant_added.notify(msg.data)
        if msg.data["ParticipantUri"] == self._account_uri and not self.session_ready.is_set():
            self.session_ready.set()

    def _handle_participant_updated(self, msg: VivoxMessage):
        participant_uri = msg.data["ParticipantUri"]
        if participant_uri in self._participants:
            participant = self._participants[participant_uri]
            participant.update(msg.data)
            self.participant_updated.notify(participant)

    def _handle_participant_removed(self, msg: VivoxMessage):
        self._remove_participant(msg.data["ParticipantUri"])

    def _remove_participant(self, participant_uri: str):
        if participant_uri in self._participants:
            participant = self._participants[participant_uri]
            del self._participants[participant_uri]
            self.participant_removed.notify(participant)

    def _global_to_region(self, pos: Vector3):
        x = pos.X - self._region_global_x * 256
        z = pos.Z + self._region_global_y * 256
        # Vivox uses a different coordinate system than SL, Y is up!
        return Vector3(x, -z, pos.Y)

    def _region_to_global(self, pos: Vector3):
        x = pos.X + self._region_global_x * 256
        y = pos.Y + self._region_global_y * 256
        return Vector3(x, pos.Z, -y)

    def _build_position_dict(self, pos: Vector3, vel: Vector3 = Vector3(0, 0, 0)) -> dict:
        return {
            "Position": {
                "X": pos.X,
                "Y": pos.Y,
                "Z": pos.Z,
            },
            "Velocity": {
                "X": vel.X,
                "Y": vel.Y,
                "Z": vel.Z,
            },
            "AtOrientation": {
                "X": "1.29938e-05",
                "Y": 0,
                "Z": -1,
            },
            "UpOrientation": {
                "X": 0,
                "Y": 1,
                "Z": 0,
            },
            "LeftOrientation": {
                "X": -1,
                "Y": 0,
                "Z": "-1.29938e-05",
            }
        }

    def _channel_info_updated(self):
        pos = self.global_pos
        if self._region_global_x is not None:
            pos = self.region_pos
        self.channel_info_updated.notify(pos)

    def _make_request_id(self):
        return str(uuid.uuid4())
